@@ -5,13 +5,16 @@ The Estimators module contains a variety of statistical estimators that can be a
 
 # TODO: Consider suggestions at https://stackoverflow.com/questions/35673895/type-hinting-annotation-pep-484-for-numpy-ndarray
 # TODO: Ensure code follows style guide: https://numpydoc.readthedocs.io/en/latest/format.html
+import multiprocessing
+
 from numba import jit
 import numpy as np
+from pathos.pools import ProcessPool
 from scipy.stats import rankdata
 import tqdm
 
 
-def kendall_tau_old(X: np.ndarray, method='A') -> np.float64:
+def kendall_tau(X: np.ndarray, method='A', n_jobs=multiprocessing.cpu_count()) -> np.float64:
     '''
     Multivariate Kendall's tau.
 
@@ -46,84 +49,22 @@ def kendall_tau_old(X: np.ndarray, method='A') -> np.float64:
     >>> kendall_tau(data)
     1.0
     '''
-
-    concordant = 0
-    discordant = 0
-
-    for i, xi in enumerate(X):
-        for j, xj in enumerate(X):
-            if i < j:
-                if (np.all(xi < xj) or np.all(xi > xj)):
-                    concordant += 1
-                else:
-                    discordant += 1
-            else:
-                continue
-    if method == 'A':
-        return (concordant - discordant) / ((X.shape[0] **2 - X.shape[0]) / 2)
-    elif method == 'B':
-        raise NotImplementedError ('Method B is not implemented yet.')
-    elif method == 'C':
-        m = np.min(X.shape)
-        result =  2 * (concordant - discordant)
-        result /= X.shape[0]**2
-        result /= (m-1) / m
-        return result
-
-def kendall_tau(X: np.ndarray, method='A') -> np.float64:
-    '''
-    Multivariate Kendall's tau.
-
-    Parameters
-    ----------
-    X : array-like
-        An m x n data matrix.
-    method : {'A', 'B', 'C'}, optional
-            The method used to account for tied points.
-            The following methods are available (default is 'a'):
-          * 'A': Original Kendall's Tau.
-          * 'B': :math:`\\tau_B = \\frac{m_c - m_d}{\\sqrt[n]{\\prod_{j=1}^{n} (m_0 - m_j)}}`
-          * 'C': :math:`\tau_C = \\frac{2 (m_c - m_d) }{m^2 \\frac{(\\max(m,n) - 1)}{\\max(m,n)}}`
-
-    Returns
-    -------
-     : np.float64
-        Multivariate Kendall's tau score
-
-    Raises
-    ------
-    NotImplementedError : Method B is not implemented yet.
-
-    References
-    ----------
-    .. [1] "Kendall's Tau.", https://en.wikipedia.org/wiki/Kendall_rank_correlation_coefficient
-
-    Examples
-    --------
-    >>> import numpy as np
-    >>> data = np.arange(100).reshape(10,10)
-    >>> kendall_tau(data)
-    1.0
-    '''
-
-    @jit(nopython=True, fastmath=True) # does fastmath actually improve speed here?
-    def grade(q):
+    @jit(nopython=True, fastmath=True)
+    def scorer(q):
         score = 0
         for j in range(X.shape[0]):
-            if np.all(X[j] < X[q]) or np.all(X[j] > X[q]):
+            if np.all(X[q] < X[j]) or np.all(X[q] > X[j]):
                 score += 1
+            elif np.all(X[q] == X[j]):
+                continue
             else:
                 score -= 1
         return score
-    score = 0
-    for i in tqdm.tqdm(range(X.shape[0])):
-        for j in range(X.shape[0]):
-            if np.all(X[i] < X[j]) or np.all(X[i] > X[j]):
-                score += 1
-            elif np.all(X[i] == X[j]):
-                continue
-            else:
-                score -= 1
+
+    pool = ProcessPool(nodes=n_jobs)
+    score = np.sum(pool.map(scorer, range(X.shape[0])))
+    pool.close()
+
     if method == 'A':
         return score / ((X.shape[0] **2 - X.shape[0]))
     elif method == 'B':
@@ -136,7 +77,7 @@ def kendall_tau(X: np.ndarray, method='A') -> np.float64:
         return result
 
 
-def grade_entropy(X: np.ndarray) -> np.float64:
+def grade_entropy(X: np.ndarray, n_jobs=multiprocessing.cpu_count()) -> np.float64:
     '''
     Computes grade entropy for a strict product order
     on the row space points.
@@ -198,8 +139,12 @@ def grade_entropy(X: np.ndarray) -> np.float64:
             if np.all(X[j] < X[q]):
                 v[j] += 1
         return v
-    for i in tqdm.tqdm(range(X.shape[0])):
-        P[:, i] = grade(i)
+
+    pool = ProcessPool(nodes=n_jobs)
+    for i, v in enumerate(pool.imap(grade, range(X.shape[0]))):
+        P[:, i] = v
+    pool.close()
+
     bins, counts = np.unique(np.sum(P, axis=0), return_counts=True)
     probs = counts / np.sum(counts)
     entropy = np.sum(probs * np.log(probs)) / (np.log(1/np.sum(counts)))
